import os
import re

from socket import error as SocketError, timeout as SocketTimeout

from urllib3.connection import HTTPConnection
from urllib3.exceptions import (NewConnectionError, ConnectTimeoutError)
from urllib3.util import connection

from dns import resolver

import logging

logger = logging.getLogger("srv_hijacker")


def resolve_ip_for_target(rrsets, target):
    for rrset in rrsets:
        if rrset.name == target:
            return rrset.items[0].address

    raise NewConnectionError("Couldn't find A record for target %s" % target)


def resolve_srv_record(old_host, resolver):
    ans = resolver.query(old_host, 'SRV')

    new_port = ans[0].port
    new_host = resolve_ip_for_target(ans.response.additional, ans[0].target)

    logger.debug("Resolved SRV record for host %s: (%s:%s)", old_host,
                 new_host, new_port)

    return new_host, new_port


original_new_conn = HTTPConnection._new_conn


def patched_new_conn(url_regex, srv_resolver):
    """
    Returns a function that does pretty much what
    `urllib3.connection.HTTPConnection._new_conn` does.

    url_regex:

    The regex to match a host against. If this regex matches the host, we
    hit the srv_resolver to fetch the new host + port
    """

    def patched_f(self):
        if re.search(url_regex, self.host):
            logger.debug("Host %s matched SRV regex, resolving", self.host)
            self.host, self.port = resolve_srv_record(self.host, srv_resolver)
        else:
            logger.debug("Host %s did not match SRV regex, ignoring",
                         self.host)

        return original_new_conn(self)

    return patched_f


def hijack(host_regex, srv_dns_host=None, srv_dns_port=None):
    """
    Usage:

    ```
    srv_hijacker.hijack(
        host_regex=r'service.consul$',
        srv_dns_host='127.0.0.1',
        srv_dns_port=8600
    )
    ```
    """
    srv_resolver = resolver.Resolver()
    if srv_dns_host:
        srv_resolver.nameservers = [srv_dns_host]
    if srv_dns_port:
        srv_resolver.port = int(srv_dns_port)

    HTTPConnection._new_conn = patched_new_conn(host_regex, srv_resolver)



import srv_hijacker
import requests
import pytest
from uuid import uuid4

CONSUL_HOST = "127.0.0.1"
CONSUL_DNS_PORT = "8600"
CONSUL_API_PORT = "8500"
CONSUL_API_URL = f"{CONSUL_HOST}:{CONSUL_API_PORT}"


def test_hijack(consul_test_service_url):
    # Before 'hijack' is run, this request must fail
    with pytest.raises(requests.exceptions.ConnectionError):
        requests.get(consul_test_service_url)

    srv_hijacker.hijack(
        host_regex=r'service.consul$',
        srv_dns_host="127.0.0.1",
        srv_dns_port="8600")

    # Now that the monkey patching is done, this should succeed
    response = requests.get(consul_test_service_url)
    assert response.status_code == 200

    # Making sure a normal requests passes even after the patch is done
    response = requests.get("https://httpbin.org/get")
    assert response.status_code == 200

    # Patching again shouldn't cause issues
    srv_hijacker.hijack(
        host_regex=r'service.consul$',
        srv_dns_host="127.0.0.1",
        srv_dns_port="8600")

    response = requests.get(consul_test_service_url)
    assert response.status_code == 200
    response = requests.get("https://httpbin.org/get")
    assert response.status_code == 200


@pytest.fixture
def consul_test_service_url():
    # Let's register a service named 'test'  to point back to consul itself
    register_service_on_consul("test", CONSUL_HOST, CONSUL_API_PORT)

    # We can now use a consul endpoint itself for testing.
    return "http://test.service.consul/v1/agent/services"


def register_service_on_consul(service_name, service_host, service_port):
    url = f"http://{CONSUL_API_URL}/v1/agent/service/register"
    response = requests.put(
        url,
        json={
            "Name": service_name,
            "Address": service_host,
            "Port": int(service_port)
        })

    assert response.status_code == 200



from setuptools import setup
from setuptools import find_packages

setup(
    name='srv_hijacker',
    version='0.0.5',
    description='Patch urllib3 to query a certain DNS server for SRV records',
    url='https://github.com/rohitpaulk/srv_hijacker',
    author='Paul Kuruvilla',
    author_email='paul.kuruvilla@shuttl.com',
    license='MIT',
    packages=find_packages(),
    classifiers=["Programming Language :: Python :: 3.7"],
    install_requires=['dnspython'])
